"""Unittests for event.py, message.py, testevent.py

Author: Thomas Low <thomas.low@cern.ch>
"""

import unittest

def suite():
	suite1 = unittest.makeSuite(MessageTestCase,'test')
	suite2 = unittest.makeSuite(EventTestCase,'test')
	suite3 = unittest.makeSuite(EventManagerTestCase,'test')
	suite4 = unittest.makeSuite(TestTestCase,'test')
	return unittest.TestSuite((suite1, suite2, suite3, suite4))


import message

class MessageTestCase(unittest.TestCase):
	
	def setUp(self):
		self.m = message.Message()
		self.m.key = 'value'
		self.m.number = '12345'
		self.m.data = 'multi\nline'
	
	def testConvertMessageToString(self):
		"""Message: Convert message to string"""
		self.assertEqual(str(self.m),
			'number:12345\nkey:value\ndetailsdata:multi\nline\n')
		
	def testConvertStringToMessage(self):
		"""Message: Convert string to message"""
		m = message.stringToMessage(str(self.m))
		self.assertEqual(m.key,'value')
		self.assertEqual(m.number, '12345')
		self.assertEqual(m.data, 'multi\nline')
		
	def testConvertStringToMessageWithNonAlphaKey(self):
		"""Message: Convert string to message with a non alpha key"""
		try:
			m = message.stringToMessage("123:abc\n")
		except ValueError:
			pass
		else:
			self.fail("ValueError expected")
		
	def testConvertDictToMessage(self):
		"""Message: Convert dict to message"""
		d = {'key': 'value', 'number' : '12345', 'data' : 'multi\nline'}
		m = message.dictToMessage(d)
		self.assertEqual(m.key,'value')
		self.assertEqual(m.number, '12345')
		self.assertEqual(m.data, 'multi\nline')
		
	def testConvertDictToMessageWithNonAlphaKey(self):
		"""Message: Convert dict with bad key (begins with digit) to message""" 
		d = {'123': 'value'}
		try:
			m = message.dictToMessage(d)
		except ValueError:
			pass
		else:
			self.fail("ValueError expected")
			
	def testConvertMessageToDict(self):
		"""Message: Convert message to dict"""
		d = message.messageToDict(self.m)
		self.assertEqual(d['key'], 'value')
		self.assertEqual(d['number'], '12345')
		self.assertEqual(d['data'], 'multi\nline')
	
	def testNotExistingArgument(self):
		"""Message: Non existing argument raises an AttributeError."""
		try:
			tmp = self.m.doesnotexist
		except AttributeError:
			pass
		else:
			self.fail("AttributeError expected")
		
	def testNonAlphaNumKey(self):
		"""Message: Assignment to a non alpha-numeric key raises a ValueError"""
		try:
			self.m.key_ads = 'value'
		except ValueError:
			pass
		else:
			self.fail("ValueError expected")
		
	def testMultiLineValue(self):
		"""Message: Multi-line-value for a non-data key raises a ValueError."""
		try:
			self.m.key2 = 'multi\nline\nvalue'
		except ValueError:
			pass
		else:
			self.fail("ValueError expected")


		
import test

class TestTestCase(unittest.TestCase):
	
	def testNameFormat(self):
		"""Test: Special format of the name attribute ^[a-zA-Z][\w]*$"""
		t = test.Test("anam2e_as34d__")
		try:
			t = test.Test("asds.sd34")
		except ValueError:
			pass
		else:
			self.fail("ValueError expected")
		try:
			t.testName = "3aas34"
		except ValueError:
			pass
		else:
			self.fail("ValueError expected")


		
import event

class EventTestCase(unittest.TestCase):
	
	def setUp(self):
		self.e = event.Event()
	
	def _eventFunction(self):
		self.check += 1
	
	def testSetupAndCallEvent(self):
		"""Event: Add/remove/call event produces right output"""
		self.check = 0
		f1 = f2 = f3 = self._eventFunction
		self.e.append(f1)
		self.e.extend([f1,f2,f3])
		self.e.remove(f1)
		self.e()
		self.assertEqual(self.check, 3)

class EventManagerTestCase(unittest.TestCase):
	
	def setUp(self):
		self.m = event.EventManager()
	
	def _eventFunction(self):
		self.check += 1
	
	def testSetupAndCallEvent(self):
		"""EventManager: Add/remove/call event produces right output"""
		self.check = 0
		f1 = f2 = f3 = self._eventFunction
		self.m.subscribe('e1',f1)
		for f in [f1,f2,f2,f3]:
			self.m.subscribe('e2',f)
		self.m.unsubscribe('e2', f2)
		self.m.call('e1')
		self.assertEqual(self.check, 1)
		self.m.call('e2')
		self.assertEqual(self.check, 4)


if __name__ == "__main__":
	unittest.main()
