"""Configuration Loading

This module provides a modification of the ConfigParser and basic functions to 
read a configuration file from different locations using the new ConfigParser.

Author: Thomas Low <thomas.low@cern.ch>
"""

import os, logging
import ConfigParser

log = logging.getLogger(__name__)


class EnvConfigParser(ConfigParser.ConfigParser):
	"""ConfigParser which is able to substitute environment variables.
	
	Additionally to the normal substitution of DEFAULT or SECTION wide values,
	this ConfigParser can also substitute environment variables (if the variable
	was not overwritten by a DEFAULT or SECTION variable. (Use: %(env_var)s for
	substitution)"""

	def get(self, section, option, raw=0, vars=None):
		try:
			val = ConfigParser.ConfigParser.get(self, section, option, raw, vars)
		except Exception:
			val = ConfigParser.ConfigParser.get(self, section, option, 1, vars)
		return self._interpolate(section, option, val, vars)
			
	def _interpolate(self, section, option, rawval, vars):
		"""Adds the additional key-value pairs to the interpolation.
		
		Also allows to use default values in %(KEY:-DEFAULT)s"""
		if not vars:
			vars = {}
		for k,v in os.environ.items():
			if not k in vars.keys():
				vars[k] = v
		
		value = rawval
		depth = ConfigParser.MAX_INTERPOLATION_DEPTH
		while depth:                    # Loop through this until it's done
			depth -= 1
			start = value.find("%(")
			if start > -1:
				end = value.find(")s",start)
				if end <= start:
					log.error("ConfigParser: no \")s\" found after \"%(\" : "+rawval)
					return ""
				rawkey = value[start+2:end]
				default = None
				key = rawkey
 				sep = rawkey.find(':-')
				if sep > -1:
					default = rawkey[sep+2:]
					key = key[:sep]            	
				if vars.has_key(key):
					value = value.replace("%("+rawkey+")s", vars[key],1)
				else:
					if default:
						value = value.replace("%("+rawkey+")s", default,1)
					else:
						log.error("ConfigParser: Key %s not found in: %s" % (rawval, key))
						return ""
			else:
				break
		if value.find("%(") != -1:
			log.error("ConfigParser: Interpolation Depth error: %s" % rawval)
			return ""
		return value


def load(name=None, homeenv=None, path=None, pathenv=None, defaults=None):
	"""Returns a ConfigParser instance after reading several config files.
	
	An argument specifies the location from where a config file will be read:
		
		NAME + HOMEENV reads a file $HOMEENV/NAME
		NAME reads a file $HOME/.NAME
		PATHENV reads a file $PATHENV
		PATH reads a file PATH
			
	If more than one file can be read, then it will be read in the same order as
	above.
	
	DEFAULTS 
		Specifies a dictionary which will be load into the "DEFAULT" section,
		so that you can use %(dict-key)s in a configuration file 
		(default: None)
	"""
	c=EnvConfigParser()
	if defaults:
		for k,v in defaults.items():
			c.set('DEFAULT', k, v)
	if name:
		if homeenv and os.environ.has_key(homeenv):
			c.read(os.environ.get(homeenv)+'/etc/'+name)
		c.read(os.environ.get('HOME')+'/.'+name)
	if pathenv:
		if os.environ.has_key(pathenv):
			c.read(os.environ[pathenv])
	if path:
		c.read(path)
	return c

		
def getSectionAsDict(cp, section):
	"""Returns a dictionary containing all pairs in this section.
	
	CP is a ConfigParser object
	SECTION is a string representing the section"""
	d = {}
	opts = cp.options(section)
	for opt in opts:
		d[opt] = cp.get(section, opt)
	return d
