"""Simple Event Manager

This module provides a simple Event Manager. You can create events and subscribe
to them easily. On calling an events calls all its subscribed functions."""

import logging
log = logging.getLogger(__name__)

class Event:
	"""Abstracts a simple event.
	
	Usage:
		e = Event()			#create an event
		e.add(f1)			#subscribe to the event.
		e.extend([f1, f2]) 	#extends event by a list of functions
		e.remove(f2) 		#remove a subscribtion.
		e(params)			#call all functions with args

	"""
	def __init__(self):
		self.targets = []
		
	def remove(self, f):
		"""Removes a function from the event. It will not be called anymore."""
		self.targets.remove(f)

	def append(self, f):
		"""Appends a single function to the event. 
		
		It will be called when calling the event."""
		self.targets.append(f)
		
	def extend(self, f):
		"""Extends the event with a list of functions."""
		self.targets.extend(f)
		
	def __call__(self, *a, **kw):
		"""Calls all subscribed functions with arbitrary arguments"""
		for f in self.targets: 
			try:
				f(*a, **kw)
			except Exception, err:
				log.exception(err)


class EventManager:
	"""Abstracts a set of events.
	
	This module can be used to manage multiple events in one class.
	
	Usage:
		mgr = EventManager().subscribe('onChange',f1)
		mgr.subscribe('onChange',f2)
		mgr.unsubscribe('onChange', f1)
		mgr.call('onChange')		
	"""
	
	def __init__(self):
		self.events = {}
	
	def subscribe(self, name, f):
		"""Subscribe to a specific event NAME with function F."""
		if self.events.has_key(name):
			e = self.events[name]
		else:
			e = self.events[name] = Event()
		e.append(f)
			
	def unsubscribe(self, name, f):
		"""Remove a function F from a certain event NAME."""
		if self.events.has_key(name):
			self.events[name].remove(f)		
		
			
	def call(self, name, *a, **kw):
		"""Call a specific event NAME with arbitrary arguments."""
		if self.events.has_key(name):
			self.events[name](*a, **kw)
