"""Logging Setup

This module provides a Colored Stream Handler (for stdout,stderr) and a function
to setup it up.

Author: Thomas Low <thomas.low@cern.ch>
"""

import sys, logging

log = logging.getLogger(__name__)


class ColoredStreamHandler(logging.StreamHandler):
	def emit(self, record):
		colored = 0
		if record.levelno > 15 and record.levelno < 25: #info
			colored = 1
			self.stream.write('\033[1;34m') #blue
		if record.levelno > 25 and record.levelno < 35: #warning
			colored = 1
			self.stream.write('\033[1;33m') #yellow
		if record.levelno > 35 and record.levelno < 45: #error
			colored = 1
			self.stream.write('\033[1;31m') #red
		if record.levelno > 45: #critical
			colored = 1
			self.stream.write('\033[1;35m') #purple
		logging.StreamHandler.emit(self, record)	
		if colored:
			self.stream.write('\033[m')
			
class ListHandler(logging.Handler):
	def __init__(self):
		logging.Handler.__init__(self)
		self.log = []
		
	def emit(self, record):
		try:
			msg = self.format(record)
			self.log.append(msg)
		except:
			self.handleError(record)
			
	def getList(self):
		self.acquire()
		log = self.log
		self.release()
		return log
          



def setupColoredStreamLogging(stream=sys.stdout, level=logging.ERROR, format=None, name=None):
	"""Setup a STREAM logger to a specific LEVEL, FORMAT and NAME.
	
	LEVEL is an integer determining the log detail (default: logging.ERROR)
	FORMAT is a logging.Formatter (default: "time level: message (name)")
	NAME is the logger's name or None (default: root)
	STREAM is the stream (default: sys.stderr)
	"""
	logger = logging.getLogger(name)
	logger.setLevel(logging.DEBUG)
	if not format:
		format = logging.Formatter(
			'%(asctime)-15s %(levelname)-4.4s %(name)s: %(message)s'
		)
	handler = ColoredStreamHandler(stream)
	handler.setLevel(int(level))
	handler.setFormatter(format)
	logger.addHandler(handler)
	
def setupListLogging(level=logging.DEBUG, format=None, name=None):
	logger = logging.getLogger(name)
	logger.setLevel(logging.DEBUG)
	if not format:
		format = logging.Formatter(
			'%(asctime)-15s %(levelname)-4.4s %(name)s: %(message)s'
		)
	handler = ListHandler()
	handler.setLevel(int(level))
	handler.setFormatter(format)
	logger.addHandler(handler)
	
	return handler.getList()
	
def setupFileLogging(filename, level=logging.ERROR, format=None, name=None):
	import logging.handlers
	logger = logging.getLogger(name)
	logger.setLevel(logging.DEBUG)
	if not format:
		format = logging.Formatter(
			'%(asctime)-15s %(levelname)-4.4s %(name)s: %(message)s'
		)
	handler = logging.handlers.RotatingFileHandler(filename)
	handler.setLevel(int(level))
	handler.setFormatter(format)
	logger.addHandler(handler)
