"""A simple key-value based Message

This module abstracts a simple key-value based message according to the Grid 
Monitoring Probe Specification at:

https://twiki.cern.ch/twiki/bin/view/LCG/GridMonitoringProbeSpecification

Author: Thomas Low <thomas.low@cern.ch>
"""

class Message:
	"""Represents a simple key-value based message with a data field.
	
	Usage:
		m = Message()
		m.key = "value"
		m.data = "multi\\nline\\nvalue"
		print m
	"""
	def __init__(self):
		self.__dict__['__values__'] = {}
		for attr in ['update', 'keys', 'values', 'items', 'get', 'has_key']:
			self.__dict__[attr] = getattr(self.__dict__['__values__'], attr)
	
	def __setattr__(self, name, value):
		"""Raises a ValueError if something is not complaint to the 
		specification.
		"""
		name = str(name)
		value = str(value)
		if name != 'data':
			if not name[:1].isalpha():
				raise ValueError("key has to begin with a letter: \"%s\"" % repr(name))
			if not name.isalnum():
				raise ValueError("key is not alpha numeric: \"%s\"" % repr(name))
			if value.find('\n') > -1:
				raise ValueError("no newline allowed")
		self.__dict__['__values__'][name] = str(value)
		return self	
		
	def __getattr__(self, name):
		try:
			return self.__dict__['__values__'][name]
		except KeyError:
			raise AttributeError("Message instance has no attribute '%s'" % name)
		
	def __repr__(self):
		return str(self.__dict__['__values__'])
		
	def __str__(self):
		return messageToString(self)
		
	def asDict(self):
		return messageToDict(self)


def messageToString(m):
	"""Converts a message to a string"""
	s = "".join(['%s:%s\n' % (str(k),str(v)) for (k,v) in m.items() 
					if k != 'data'])
	d = ""
	if m.has_key('data'):
		d = "detailsdata:%s\n" % m.data
	return s+d


def stringToMessage(s):
	"""Converts a string into a Message.
	
	Raises a ValueError if something is not compliant to the specification.
	"""
	m = Message()
	lines = s.split("\n")
	fdata = 0
	data = []
	count = 0
	for line in lines:
		count += 1
		if count == len(lines):
			if line != '':
				raise ValueError("no empty line at the end", repr(s))
		elif fdata:
			data.append(line)
		else:
			if line == '':
				raise ValueError("no empty line allowed here", repr(s))
			if line[:12] == "detailsdata:":
				fdata = 1
				data.append(line[12:])
			else:
				kv = line.split(":",1)
				if len(kv) < 2:
					raise ValueError("no separator \":\" found", repr(line))
				setattr(m, kv[0].strip(), kv[1].strip())
	if data:
		m.data = "\n".join(data)
	return m
	
	
def dictToMessage(d):
	"""Converts a dict into a Message.
	
	Raises a ValueError if something is not compliant to the specification."""
	m = Message()
	for k,v in d.items(): setattr(m,k,v)
	return m


def messageToDict(m):
	"""Converts a message into a dict"""
	return m.__dict__['__values__'].copy()
	
