"""TestManager for TestProducers and TestConsumers

Manages a set of producer- and consumer-classes.

Author: Thomas Low <thomas.low@cern.ch>
"""

import logging
log = logging.getLogger(__name__)

from gcm.common.event import Event


class TestManager:
	"""Manages a producer- and consumer-classes with an event."""
	
	def __init__(self):
		"""Init Testmanager."""
		self.producer = []
		self.consumer = []
		self.all = []
		self.event = Event()
		
		
	def addProducer(self, module, args):
		"""Adds a producer to the TestManager.
		
		Each Producer's run method will be called with the global event.
		MODULE definies the path to the python class.
		ARGS definies a dictionary of arguments for the Producers init method.
		"""
		try:
			obj = dynamic_import(module)
			inst = obj(**args)
			self.producer.append(inst)
			self.all.append(inst)
		except Exception, err:
			log.warning("%s could not be initialized: %s" % (str(module),str(err)))
			log.exception(err)		
	
	def addConsumer(self, module, args):
		"""Adds a consumer to the TestManager.
		
		Each Consumer's consume method will be called, when a Producer triggers
		the global event.
		MODULE definies the path to the python class.
		ARGS definies a dictionary of arguments for the Producers init method.
		"""
		try:
			obj = dynamic_import(module)
			inst = obj(**args)
			self.consumer.append(inst)
			self.all.append(inst)
			self.event.append(inst.consume)
		except Exception, err:
			log.warning("%s could not be initialized: %s" % (str(module),str(err)))
			log.exception(err)
	
	def doSetup(self):
		"""Setup all modules by calling their setup methods."""
		for c in self.all:
			try:
				log.info("setup %s" % str(c))
				c.setup()
			except Exception, err:
				log.warning("%s setup failed: %s" % (str(c),str(err)))
				log.exception(err)
	
	def doRun(self):
		"""Run all producers by calling their run methods."""
		for c in self.producer:
			try:
				log.info("run %s" % str(c))
				c.run(self.event)
			except Exception, err:
				log.warning("%s failed while running: %s" % (str(c),str(err)))
				log.exception(err)		
	
	def doCleanup(self):
		"""Cleanup all modules by calling their cleanup modules.
		
		This happens in the reverse order of doSetup().
		"""
		reverse = []
		for a in self.all:
			reverse.insert(0,a)
		for c in reverse:
			try:
				log.info("clean up %s" % str(c))
				c.cleanup()
			except Exception, err:
				log.warning("%s cleanup failed: %s" % (str(c),str(err)))
				log.exception(err)


def dynamic_import(path):
	"""Dynamically import a module specified by PATH."""
	modules = str(path).strip().split(".")
	if len(modules) == 0:
		return None
	elif len(modules) == 1:
		exec "obj = "+path
		return obj
	else:
		exec "import "+".".join(modules[:-1])
		exec "obj = "+path
		return obj
