"""PyAMQ Consumer

This module represents an asynchronous activeMQ Consumer written in Python.

Author: Thomas Low <thomas.low@cern.ch>
"""

import sys, time, threading, logging
from lib import brokermgr, util

log = logging.getLogger(__name__)

class Consumer(threading.Thread):
	"""Represents an asynchronous activeMQ Consumer.
	
	Each time a message is received, the datahandler will be called with it."""
	
	def __init__(self, broker, random, destination, datahandler):
		"""Initializes the consumer for a destination.
		
		BROKER is a list of broker URIs e.g. ['stomp://host:port/?timeout=3&retries=3']
		RAMDOM==0 means, that broker are choosen sequentially from the list.
		DESTINATION is the activeMQ destination of the topic/queue
		DATAHANDLER is a function with one argument (the message)
		"""
		self.stopped = 0
		self.lock = threading.Lock()
		self.event = threading.Event()
		
		self.datahandler = datahandler
		self.dataqueue = []
		
		self.brokermgr = brokermgr.BrokerManager(util.getUrlsAsDicts(broker),
							1, random, self._event_handler, destination)
		self.brokermgr.start()
		threading.Thread.__init__(self)
		self.start()


	def __del__(self):
		"""Stops the everything immediately."""
		self.stop(0)
	
	
	def _event_handler(self, event, data = None):
		"""Will be called by the BrokerManager with the message."""
		if event == 'MESSAGE_RECEIVED':
			self.lock.acquire()
			self.dataqueue.append(data)
			self.lock.release()
			self.event.set()
		elif event == 'EXIT':
			self.stopped = 1
			self.event.set()
	
			
	def stop(self, timeout=0.5):
		"""Stops the consumer within maximal TIMEOUT seconds."""
		self.brokermgr.stop(timeout)
		self.stopped = 1
		self.event.set()
		self.join()
	
	
	def waitUntilKeyboardInterrupt(self):
		import signal
		try:
			signal.pause()
		except KeyboardInterrupt:
			pass
	
	
	def run(self):
		alive = 1
		while alive:
			self.lock.acquire()
			while self.dataqueue and not self.stopped:
				tmp = self.dataqueue.pop()
				self.lock.release()	
				try:
					self.datahandler(tmp)
				except Exception, err:
					log.exception(err)
				self.lock.acquire()
			self.lock.release()		
			self.event.wait()
			self.event.clear()
			
			if self.stopped:
				alive = 0


"""Test the consumer if not imported"""
if __name__ == '__main__':
	#config logging
	logging.getLogger('').setLevel(logging.INFO)
	console = logging.StreamHandler(sys.stdout)
	console.setLevel(logging.DEBUG)
	console.setFormatter(logging.Formatter(
		'%(relativeCreated)8d %(levelname)-8s: %(message)s (%(name)s)'
	))
	logging.getLogger('').addHandler(console)
	#config logging

	broker 	= ["stomp://gridmsg001.cern.ch:6163","stomp://gridmsg002.cern.ch:6163"]
	dest	= "/topic/testtopic"
	
	def handler(data):
		print data

	log.info("program started")
	c = Consumer(broker, 0, dest, handler)
	c.waitUntilKeyboardInterrupt()
	c.stop()
	log.info("program stopped")		
	
