"""broker.py"""

import sys, os, time, logging
import threading, asyncore, socket
import re

log = logging.getLogger(__name__)

ip_pattern = re.compile("\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}")

class Broker(asyncore.dispatcher):

	def __init__(self, info, event_handler):
		self.lock 			= threading.Lock()
		self.info 			= info
		self.event_handler 	= event_handler
		self.busy 			= 1
		self.closed			= 0
		self.subscription 	= None
		
		self.write_buffer 	= ''			
		self.read_buffer 	= ''			
		self.connected		= 0
		self.connecting		= 0
		
		asyncore.dispatcher.__init__(self)
			
		if ip_pattern.match(self.info['host']):
			self._getIpHandler(self.info['host'])
		else:		
			DNSThread(self.info['host'], self._getIpHandler)
	
	
	def _getIpHandler(self, host):
		if not ip_pattern.match(host):
			self.kill()
		else:
			self.info['ip'] = host
			self.lock.acquire()
			if not self.closed:
				self.event_handler(self, 'BROKER_INITIALIZED')
			self.lock.release()
	
									
	def _connect(self):
		log.debug(str(self)+" connecting to ("+str(self.info['ip'])+","+str(self.info['port'])+")")
		try:	
			self.connecting = 1
			self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
			self.connect((self.info['ip'], int(self.info['port'])))	
		except Exception, e:
			self.kill()
			log.warning(str(self)+" connection failed: "+str(sys.exc_info()[0])+" ("+str(e[1])+")")		
	
			
	def __str__(self):
		if self.info['port']:
			return str(self.info['host'])+":"+str(self.info['port'])
		else:
			return str(self.info['host'])
	
	
	def writable(self):
		return self.connecting or len(self.write_buffer) or self.closed
	
			
	def handle_connect(self):
		self.connecting = 0
		self.lock.acquire()
		self.connected = 1
		self.lock.release()
		self.write_buffer += self.doConnect()
		

	def handle_read(self):
		pass
		try: 
			tmp = self.recv(8192)
			self.read_buffer += tmp
		except Exception, e:
			log.warning(str(self)+" reading failed: "+str(sys.exc_info()[0])+" ("+str(e[1])+")")
			self.kill()
			return
			
		#print '\033[1;32m'+"--- msg received via",self.info['host'],"at",time.time(),"---"
		#print repr(self.read_buffer)
		#print "-"*80+'\033[0m'		
		
		decoded = self.decodeMessage(self.read_buffer)
		if decoded:
			self.read_buffer = self.read_buffer[decoded:]
	
	
	def handle_write(self):
		if self.write_buffer:
		
			#print '\033[1;34m'+"---","send msg via",self.info['host'],"at",time.time(),"---"
			#print repr(self.write_buffer)
			#print "-"*80+'\033[0m'
		
			sent = 0
			try:
				sent = self.send(self.write_buffer)
			except Exception, e:
				log.warning(str(self)+" writing failed: "+str(sys.exc_info()[0])+" ("+str(e[1])+")")
				self.kill()
				return
			self.write_buffer = self.write_buffer[sent:]
	
	
	def handle_close(self):
		self.kill(1)
	

	def disconnect(self):
		if self.connected:
			if self.subscription:
				self.write_buffer += self.doUnsubscribe(self.subscription)
				self.subscription = None
			self.write_buffer += self.doDisconnect()
			if not self.write_buffer: self.stop()
		else:
			self.kill()
	
	
	def stop(self):
		"""close socket gently by closing the write-pipe first"""
		if self.connected:
			try:
				self.shutdown(1)
			except:
				log.debug(str(self)+" already closed")
		else:
			self.kill()
	
	
	def kill(self, serverside=0):
		"""kill broker by closing the socket"""
		if self.connected:
			self.write_buffer = ''
			self.read_buffer = ''
			try:
				self.shutdown(2)
			except:
				log.debug(str(self)+" r/w pipes already shut down")
		try:
			self.close()	
		except:
			log.debug(str(self)+" already closed")
		self.lock.acquire()
		if not self.closed:
			self.connected = 0
			self.closed = 1
			if serverside:
				self.event_handler(self, 'BROKER_CLOSED')
			else:
				self.event_handler(self, 'BROKER_KILLED')
			log.debug(str(self)+" disconnected")
		self.lock.release()
	
	
	def _onConnected(self):
		self.lock.acquire()
		self.busy = 0
		self.lock.release()
		self.event_handler(self,'BROKER_CONNECTED')
	
	
	def _onMsgSuccess(self, busy=1):
		if busy:
			self.lock.acquire()
			self.busy = 0
			self.lock.release()
		self.event_handler(self, 'MESSAGE_SENT')
	
			
	def _onSubscribeSuccess(self):
		self.lock.acquire()
		self.busy = 0
		self.lock.release()
		self.event_handler(self, 'SUBSCRIBED')
	
	
	def _onMsgReceive(self, data):
		self.event_handler(self, 'MESSAGE_RECEIVED', data)
	

	def isConnected(self):
		return self.connected
	
											
	def isBusy(self):
		return self.busy
		
		
	def isClosed(self):
		return self.closed
		
	
	def sendMessage(self, data):
		if not self.busy:
			self.lock.acquire()
			try:
				self.busy = 1
				self.write_buffer += self.encodeMessage(data)
			finally:	
				self.lock.release()
			return 1
		else:
			return 0
			
	def subscribe(self, destination):
		if not self.busy:
			self.lock.acquire()
			try:
				self.busy = 1
				self.subscription = destination
				self.write_buffer += self.doSubscribe(destination)
			finally:
				self.lock.release()
			return 1
		else:
			return 0
			
class DNSThread(threading.Thread):
	
	def __init__(self, hostname, handler):
		self.hostname = hostname
		self.handler = handler
		threading.Thread.__init__(self)
		self.setDaemon(1)
		self.start()
	
	def run(self):
		try:
			host = socket.gethostbyname(self.hostname)
		except Exception, err:
			host = ""
			log.warning(str(self)+" could not resolve hostname \"%s\": %s" % (self.hostname, err))
		try:
			self.handler(host)
		except Exception, err:
			log.exception(err)
			
if __name__ == "__main__":
	#config logging
	logging.getLogger('').setLevel(logging.DEBUG)
	console = logging.StreamHandler(sys.stdout)
	console.setLevel(logging.DEBUG)
	console.setFormatter(logging.Formatter(
		'%(relativeCreated)8d %(levelname)-8s: %(message)s (%(name)s)'
	))
	logging.getLogger('').addHandler(console)
	#config logging

	def handler(ip): print ip
	thread = DNSThread("lxplus240.cern.ch", handler)
	time.sleep(1)
