"""brokermgr.py"""

import sys, os, time, logging
import threading, asyncore

import broker

from pyamq.protocols import stomp, rest

log = logging.getLogger('pyamq.lib.brokermgr')

protocolmap = {	'stomp'	: stomp.StompBroker, 'http' : rest.RestBroker}

class BrokerManager(threading.Thread):
	"""Encapsulates a list of brokers as it would be one (asynchroniously).
	
	Usage:
		mgr = BrokerManager(broker, maxbroker, random, event, handler)
		mgr.start()
		...
		mgr.sendMessage(message)
		...
		--> event handler called with incoming message
		...
		mgr.stop()
		
	Input:
		broker 
			list of dictionaries with keys:
			protocol, host, port, path, timeout, retries
		
		maxbroker		
			number of simulatanious connections
		
		random
			if false, broker will be extracted sequentially from broker list
			
		handler
			will be called if a message arrives
	"""

	def __init__(self, broker, maxbroker, random, event_handler, subscr = None):
		self.stopped		= 0
		self.stop_timeout	= 0.5
		
		self.asyncore_lock	= threading.Lock()
		self.lock			= threading.Lock()
		self.event			= threading.Event()
		self.event_lock		= threading.Lock()
		self.event_handler 	= event_handler
		
		self.broker_active  = []
		self.broker_info 	= broker
		self.broker_max 	= maxbroker
		self.broker_random	= random
		self.broker_events	= []
		
		self.subscription	= subscr
		
		self.msg_id			= 1
		self.msg_pending	= []
		self.msg_active		= []
		self.msg_sent		= []
		
		for broker in self.broker_info:
			if not broker.has_key('retries'):
				broker['retries'] = 4
			else:
				broker['retries'] = int(broker['retries'])
			if not broker.has_key('timeout'):
				broker['timeout'] = 1
			else:
				broker['timeout'] = int(broker['timeout'])
			broker['suspended'] = None
		
		threading.Thread.__init__(self)
		

	def sendMessage(self, message):
		"""Sends a message consisting of a dictionary (dest, header, body)."""
		if self.stopped:
			return 0
		
		self.lock.acquire()
		self.msg_pending.append(
			{'id':self.msg_id, 'data':message, 'assigned':None}
		)
		self.msg_id += 1
		self.lock.release()
		self.event.set()
	
	
	def getPendingMessages(self):
		"""Returns a list of ids of messages not sent yet"""
		self.lock.acquire()
		getdata = lambda msg: msg['data']
		tmp = []
		tmp.extend(map(getdata,self.msg_pending))
		tmp.extend(map(getdata,self.msg_active))
		self.lock.release()
		return tmp
	
	
	def getSentMessages(self):
		"""Returns a list of ids of messages which were sent since last call"""
		self.lock.acquire()
		getdata = lambda msg: msg['data']
		tmp = []
		tmp.extend(map(getdata,self.msg_sent))
		self.lock.release()
		return tmp
					
							
	def stop(self, timeout):
		"""Stops the BrokerManager in maximal "timeout" seconds"""
		self.stopped = 1
		self.stop_timeout = timeout
		self.event.set()
		try:
			self.join()
		except:
			log.warning("brokermgrThread joined though thread not started")
	
	
	### now internal functions ###
		
	def _broker_event_handler(self, broker, event, data=None):
		self.event_lock.acquire()
		self.broker_events.append({'broker':broker, 'event':event, 'data':data})
		self.event_lock.release()
		self.event.set()
			
	
	def run(self):
		alive = 1
		lock_acquired = 0
		loop = AsyncoreThread(self.asyncore_lock, self._checkBroker)
		loop.start()
		try:
			while alive:
				self.lock.acquire()
				lock_acquired = 1
				self._processBrokerEvents()
				self._addBroker()
				self._assignSubscription()
				self._assignMessages()
				self.lock.release()
				lock_acquired = 0
				
				self.event.wait()
				self.event.clear()
				if self.stopped:
					alive = 0
		finally:
			if lock_acquired:
				self.lock.release()
			self._shutdownBroker(self.stop_timeout)
			loop.stop()
			try:
				self.event_handler('EXIT')
			except Exception, err:
				log.exception(err)
		
		
	def _processBrokerEvents(self):
		self.event_lock.acquire()
		try:
			while self.broker_events:
				e = self.broker_events.pop(0)
				
				broker = None
				for b in self.broker_active:
					if id(b['obj']) == id(e['broker']):
						broker = b
						break
				
				if broker:
					#log.debug(str(broker['obj'])+" "+str(e['event']))
					#broker ready to connect
					if e['event'] == 'BROKER_INITIALIZED':
						broker['time'] = time.time()	
						self.event_lock.release()
						#self.asyncore_lock.acquire()
						broker['obj']._connect()
						#self.asyncore_lock.release()
						self.event_lock.acquire()
					#broker connected
					if e['event'] == 'BROKER_CONNECTED':
						broker['time'] = time.time()	
					#broker closed
					elif e['event'] == 'BROKER_CLOSED':
						self.broker_active.remove(broker)
						log.debug("%s closed (%d retries" % (str(broker['obj']), broker['info']['retries']))
						del broker['obj']
						if broker['msg']:
							self.msg_active.remove(broker['msg'])
							broker['msg']['assigned'] = None
							self.msg_pending.append(broker['msg'])
							broker['info']['suspended'] = broker['time']+broker['info']['timeout']
							if broker['info']['retries'] > 0:
								broker['info']['retries'] -= 1
								self.broker_info.append(broker['info'])
							elif broker['info']['retries'] == -1:
								self.broker_info.append(broker['info'])
						else:
							broker['info']['suspended'] = 0
							self.broker_info.insert(0,broker['info'])
					#broker killed on error		
					elif e['event'] == 'BROKER_KILLED':
						self.broker_active.remove(broker)
						log.debug("%s killed (%d retries" % (str(broker['obj']), broker['info']['retries']))
						del broker['obj']
						if broker['msg']:
							self.msg_active.remove(broker['msg'])
							broker['msg']['assigned'] = None
							self.msg_pending.append(broker['msg'])
						#retry connect?
						broker['info']['suspended'] = broker['time']+broker['info']['timeout']
						if broker['info']['retries'] > 0:
							broker['info']['retries'] -= 1
							self.broker_info.append(broker['info'])
						elif broker['info']['retries'] == -1:
							self.broker_info.append(broker['info'])
					#message sent
					elif e['event'] == 'MESSAGE_SENT':
						self.msg_active.remove(broker['msg'])
						self.msg_sent.append(broker['msg'])
						try:
							self.event_handler('MESSAGE_SENT', broker['msg']['data'])
						except Exception, err:
							log.exception(err)
						log.debug("%s message %s sent in %.4f seconds" \
							% (str(broker['obj']), broker['msg']['id'], time.time()-broker['msg']['assigned']))
						broker['msg'] = None
						broker['time'] = time.time()
					# message received
					elif e['event'] == 'MESSAGE_RECEIVED':
						broker['time'] = time.time()
						try:
							self.event_handler('MESSAGE_RECEIVED', e['data'])
						except Exception, err:
							log.exception(err)
				else:
					log.warning("brokerevent from unknown broker "+str(e['broker'])+": "+str(e['event']))
		finally:
			self.event_lock.release()
	
	
	def _checkBroker(self):
		self.lock.acquire()
		#check for timeouts
		timeout_broker = [b for b in self.broker_active \
			if b['time']+b['info']['timeout'] < time.time() \
				and b['obj'].isBusy() and not b['obj'].isClosed()]
		#check for suspended brokers which can be retried
		suspended_broker = [b for b in self.broker_info if b['suspended'] \
							and b['suspended'] < time.time()]
		for b in suspended_broker:
			b['suspended'] = None		
		self.lock.release()
		
		#do actions
		for b in timeout_broker:
			#log.warning(str(b['obj'])+" timeout")
			b['obj'].kill()
					
		if timeout_broker or suspended_broker:
			self.event.set()
				
	
	def _addBroker(self):
		"""Adds new broker to the active broker list"""
		if len(self.broker_active) == self.broker_max:
			return
		#add and initialize new broker
		new_broker = [b for b in self.broker_info if not b['suspended']]
		while len(self.broker_active) < self.broker_max and len(new_broker):
			if self.broker_random:
				info = new_broker.pop(int((time.time()*100) % len(new_broker)))
			else:
				info = new_broker.pop(0)
			self.broker_info.remove(info)
			try:
				b = {
					'obj'	: protocolmap[info['protocol']](info, self._broker_event_handler),
					'time'	: time.time(),
					'info'	: info,
					'msg'	: None,
					'subs'	: None
				}
				self.broker_active.append(b)
				#log.info(str(b['obj'])+" added")
			except Exception, err:
				log.exception(err)
				log.error("broker add exception: %s" % str(sys.exc_info()[0]))
				log.error("probably wrong protocol: %s" % str(info['protocol']))
		
		#if no broker left to add
		if not self.broker_active and not self.broker_info:
			self.stopped = 1
			self.event.set()
			log.critical("no broker left")
			
		
	def _assignSubscription(self):
		if not self.subscription:
			return
		if not [b for b in self.broker_active if b['subs']]:
			idle = [b for b in self.broker_active if not b['msg'] and not b['subs']]
			while idle:
				broker = idle.pop(int((time.time()*100) % len(idle)))
				if broker['obj'].subscribe(self.subscription):
					broker['subs'] = 1
					log.debug("%s subscribed to %s" % (str(broker['obj']),self.subscription))
					break
		
		
	def _assignMessages(self):
		if not self.msg_pending:
			return
		idle = [b for b in self.broker_active if not b['msg'] and not b['subs']]
		while self.msg_pending and idle:
			broker = idle.pop(int((time.time()*100) % len(idle)))
			msg = self.msg_pending[0]
			if broker['obj'].sendMessage(msg['data']):
				msg['assigned'] = time.time()
				broker['msg'] = msg
				broker['time'] = time.time()
				self.msg_active.append(msg)
				self.msg_pending.pop(0)
				
	
	def _shutdownBroker(self, timeout=0.5):
		"""Shutdown all broker, first disconnect and than close the socket"""
		if not self.broker_active:
			return
		#disconnect
		for b in self.broker_active:
			b['obj'].disconnect()
		endtime = time.time()+(float(timeout)/2.0)
		#log.debug("disconnect for %f seconds", endtime-time.time())
		while endtime - time.time() > 0 \
				and [b for b in self.broker_active if not b['obj'].isClosed()]:
			self.event.wait(endtime - time.time())
		#stop
		for b in self.broker_active:
			b['obj'].stop()
		endtime = time.time()+(float(timeout)/2.0)
		while endtime - time.time() > 0 \
				and [b for b in self.broker_active if not b['obj'].isClosed()]:
			self.event.wait(endtime - time.time())
		#kill
		for b in self.broker_active:
			b['obj'].kill()
		

class AsyncoreThread(threading.Thread):
	def __init__(self, lock, check):
		self.lock = lock
		self.check = check
		self.alive = 1
		threading.Thread.__init__(self)
	
	def run(self):
		while self.alive:
			#self.lock.acquire()
			try:
				if asyncore.socket_map:
					asyncore.poll(0.01)
				else:
					time.sleep(0.01)
			except:
				log.critical("asyncore exception: %s" % str(sys.exc_info()[0]))
			self.check()
			
	
	def stop(self):
		self.alive = 0
		try:
			self.join()
		except:
			log.warning("asyncoreThread stopped though thread not started")
