"""Process Manager

This module contains a Process Manager which can for subprocesses and capture
their stdout, stderr. It also can force a process to quit after a certain
amount of time.

Author: Thomas Low <thomas.low@cern.ch>
"""

import os, sys, time, logging
import select, signal, threading

log = logging.getLogger(__name__)

class ProcessManager(threading.Thread):

	def __init__(self, max_processes=None):
		"""Inits the ProcessManager class.
		
		MAX_PROCESSES is the maximum count of simultaneous subprocesses.
		"""
		self.procs_pending 	= []
		self.procs_active 	= []
		self.procs_max		= max_processes
		
		self.stopping = 0
		self.stopGently = 0
		self.proc_event_r, self.proc_event_w = os.pipe()
		self.wait_event_r, self.wait_event_w = os.pipe()
		self.lock = threading.Lock()
		
		threading.Thread.__init__(self)
		self.start()
		
		
	def wait(self, timeout):
		"""Blocks until all commands are excecuted or the timeout exceeded."""
		endtime = time.time() + timeout
		while not self.stopping and self.procs_pending + self.procs_active:
			timeleft = endtime - time.time()
			if timeleft > 0:
				try:
					triggered = select.select([self.wait_event_r],[],[], timeleft)
					if triggered[0]:
						os.read(self.wait_event_r,32)
				except:
					return 1
			else:
				return 0
		return 0
			
			
	def stop(self):
		"""Stops the ProcessManager immediately."""
		if not self.stopping:
			#log.info("stopping ProcessManager")
			self.stopping = 1
			try:
				os.write(self.proc_event_w, "\n")
				os.write(self.wait_event_w, "\n")
			except:
				log.info("file descriptor already closed")
			try:
				self.join()
			except:
				log.warning("ProcessManager joining error")
	
	
	def stopWhenFinished(self):
		"""Stops the ProcessManager as soon as no programs are still running."""
		if not self.stopping:
			self.stopGently = 1
			try:
				os.write(self.proc_event_w, "\n")
			except:
				log.info("file descriptor already closed")
	
			
	def execute(self, cmd, handler, timeout, stdout=None):
		"""Executes a command.
		
		 HANDLER is a function which will be called with a dictinary of data.
		 TIMEOUT is the maximal time in seconds a process is allowed to live.
		 STDOUT is a handler which will be called with the stdout, stderr while
		 	the process is still running. (e.g. to output logging)
		"""
		if not self.stopping and not self.stopGently:
			self.lock.acquire()
			self.procs_pending.append(Subprocess(cmd, handler, timeout, stdout))
			self.lock.release()
			try:
				os.write(self.proc_event_w, "\n")
			except:
				log.info("file descriptor already closed")
		
		
	def run(self):
		"""Handles all the functionalities mentioned above."""
		alive = 1
		try:
			while alive:
				# add pending processes
				self.lock.acquire()
				while self.procs_pending and (not self.procs_max > 0 \
						or len(self.procs_active) < self.procs_max):
					proc = self.procs_pending.pop(0)
					#if proc.getTimeLeft() > 0:
					self.procs_active.append(proc)
						#log.debug(str(proc)+" started")
					proc.start()
					#else:
						#proc not started (timeout before start)")
					#	self.lock.release()
					#	proc.callHandler()
					#	self.lock.acquire()
					#	os.write(self.wait_event_w, "\n")
					#	os.write(self.proc_event_w, "\n")
					#	del proc
				self.lock.release()
				
				# create wait list and do select
				wait_objects = [p.getWrapper() for p in self.procs_active]
				waitfor = [x for tupls in wait_objects for x in tupls]
				waitfor.append(self.proc_event_r)
				
				proc_timeouts = [p.getTimeLeft() for p in self.procs_active]
				
				if proc_timeouts:
					triggered = select.select(waitfor,[],[], max(0,min(proc_timeouts)))
				else:
					triggered = select.select(waitfor,[],[])
				
				#manage output of subprocesses			
				for wrapper in triggered[0]:
					if wrapper == self.proc_event_r:
						os.read(self.proc_event_r,32)
					else:
						proc = wrapper.getSubprocess()
						proc.readData(wrapper.fileno())	
									
				for proc in [p for p in self.procs_active if p.getTimeLeft() < 0]:
					proc.kill()
				
				for proc in [p for p in self.procs_active if p.hasFinished()]:
					proc.close()
					#log.debug(str(proc)+" finished with: "+str(proc.exitcode))
					proc.callHandler()
					self.lock.acquire()
					self.procs_active.remove(proc)
					self.lock.release()
					del proc
					os.write(self.wait_event_w, "\n")
					
				if self.stopGently and not self.procs_active and not self.procs_pending:
					self.stopping = 1
					
				if self.stopping:
					alive = 0
		finally:
			try:
				os.write(self.wait_event_w, "\n")
			except:
				log.info("file descriptor already closed")
			for fd in [self.proc_event_r, self.proc_event_w, self.wait_event_r, 
						self.wait_event_w]:
				try:
					os.close(fd)
				except:
					log.info("file descriptor already closed")
			#kill all active subprocesses
			for proc in self.procs_active:
				proc.kill()
				proc.close()
				proc.callHandler()
			for proc in self.procs_pending:
				proc.callHandler()

		
class Subprocess:
	"""Represents a subprocess."""
	
	def __init__(self, cmd, handler, timeout, stdout_handler=None, bufsize=1):
		"""Inits a subprocess.
		
		CMD is the command to execute.
		HANDLER is the function which will be called after the process has quit.
		TIMEOUT is the time in seconds a process is allowed to run.
		STDOUT_HANDLER is an optional handler which will be called while the
			process is running (with its stdout and stderr).
		BUFSIZE is the buffer size for the communication between the
			ProcessManager and the subprocesses. (Default is line-based).
		"""
		self.cmd		= cmd
		self.handler	= handler
		self.stdout_h	= stdout_handler
		self.stdout 	= []
		self.stderr 	= ['not started']
		self.exitcode	= None
		self.starttime	= 0
		self.endtime	= 0
		
		self.pid		= None
		self.stdout_eof = 0
		self.stderr_eof	= 0
		self.fo_stdin	= None
		self.fo_stdout 	= None
		self.fo_stderr	= None	
		self.bufsize	= bufsize
		self.started	= 0
		self.closed		= 0
		self.killed 	= 0
		self.killtime	= 0
		self.timeout	= timeout
							
	def __del__(self):
		"""Waits for the subprocess to exit."""
		self.close()
		
	def __str__(self):
		return os.path.basename(self.cmd)[:20]
		
	def start(self):
		"""Starts the subprocess by forking it."""
		self.killtime = time.time() + self.timeout
		self.starttime = time.time()
		p2cr, p2cw = os.pipe()
		c2pr, c2pw = os.pipe()
		errr, errw = os.pipe()
		try:
			self.pid = os.fork()
		except Exception, err:
			log.exception(err)
			self.starttime = 0
			self.killtime = 0
			return
		
		if self.pid == 0:
			os.setpgrp()
			os.nice(1)
			os.dup2(p2cr, 0)
			os.dup2(c2pw, 1)
			os.dup2(errw, 2)
			os.close(p2cr)
			os.close(p2cw)
			os.close(c2pr)
			os.close(c2pw)
			os.close(errr)
			os.close(errw)
			param = ['/bin/sh', '-c', self.cmd]
			try:
				os.execvp(param[0], param)
			finally:
				os._exit(1)
		else:
			os.close(p2cr)
			os.close(c2pw)
			os.close(errw)
			self.fo_stdin = p2cw
			self.fo_stdout = c2pr
			self.fo_stderr = errr
			self.started = 1
			self.stderr = []
		
	def readData(self, fo):
		"""Reads the data from the subprocess and calls the handler."""
		if fo == self.fo_stdout:
			tmp = os.read(self.fo_stdout, 32)
			if tmp != '':
				self.stdout.append(tmp)
				if self.stdout_h:
					try:
						self.stdout_h(tmp)
					except Exception, err:
						log.exception(err)
			else:
				self.stdout_eof = 1
		elif fo == self.fo_stderr:
			tmp = os.read(self.fo_stderr, 32)
			if tmp != '':
				self.stderr.append(tmp)
				if self.stdout_h:
					try:
						self.stdout_h(tmp)
					except Exception, err:
						log.exception(err)
			else:
				self.stderr_eof = 1
		if self.stdout_eof and self.stderr_eof:
			self.endtime = time.time()
		
	def getFileObjects(self):
		"""Returns the file objects for stdin, stdout, stderr."""
		return self.fo_stdin, self.fo_stdout, self.fo_stderr
		
	def getWrapper(self):
		"""Returns a Wrapper class for a file object."""
		result = []
		if not self.stdout_eof:
			result.append(_OutWrapper(self))
		if not self.stderr_eof:
			result.append(_ErrWrapper(self))	
		return result
				
	def hasFinished(self):
		"""Returns whether the subprocess is already dead."""
		return (self.stdout_eof and self.stderr_eof) or self.killed
		
	def getTimeLeft(self):
		"""Returns the time a subprocess has left."""
		return self.killtime - time.time()
		
	def callHandler(self):
		"""Calls the handler for the subprocess with a dictionary of data."""
		if self.handler:
			try:
				ec = os.WEXITSTATUS(int(self.exitcode))
				es = os.WTERMSIG(int(self.exitcode))
			except:
				ec = ''
				es = ''
			try:
				self.handler(self.cmd, {
					'cmd'		: self.cmd,
					'stdout'	: ''.join(self.stdout),
					'exitcode'	: ec,
					'signal'	: es,
					'stderr'	: ''.join(self.stderr),
					'starttime'	: self.starttime,
					'endtime'	: self.endtime		
					})
			except Exception, err:
				log.exception(err)
		
	def kill(self):
		"""Just kills a subprocess by sending a SIGKILL."""
		log.debug(str(self)+" killing")
		try:
			os.kill(-self.pid, signal.SIGKILL)
		except:
			log.warning("%s process already dead" % self.cmd)
		self.killed = 1
		self.endtime = time.time()
		self.stderr.append('killed after timeout of %.1f seconds' % self.timeout)
		
	def close(self):
		"""Waits for a subprocess to die."""
		if self.started and not self.closed:
			self.closed = 1
			try:
				os.close(self.fo_stdin)
				os.close(self.fo_stdout)
				os.close(self.fo_stderr)
				pid, sts = os.waitpid(self.pid, 0)
				if pid == self.pid:
					self.exitcode = sts
			except:
				log.error("process close exception: %s" % str(sys.exc_info()[0]))
		

class _FOWrapper:
	def __init__(self, subprocess):
		self.subprocess = subprocess
	def getSubprocess(self):
		return self.subprocess
			
class _OutWrapper(_FOWrapper):
	def fileno(self):
		return self.subprocess.getFileObjects()[1]
		
class _ErrWrapper(_FOWrapper):
	def fileno(self):
		return self.subprocess.getFileObjects()[2]



"""Test the processmgr if not imported"""
if __name__ == '__main__':
	#config logging
	logging.getLogger('').setLevel(logging.INFO)
	console = logging.StreamHandler(sys.stdout)
	console.setLevel(logging.DEBUG)
	console.setFormatter(logging.Formatter(
		'%(relativeCreated)8d %(levelname)-8s: %(message)s (%(name)s)'
	))
	logging.getLogger('').addHandler(console)
	#config logging
	
	def handler(cmd,result):
		log.info(cmd+" handled with exitcode: "+str(result['exitcode']))

	def stdout_h(data):
		print data,
			
	mgr = ProcessManager(3)
	for i in range(70,99):
		mgr.execute("echo 0."+str(i), handler, 5, stdout_h)
	mgr.stopWhenFinished()
	log.info("Stop when finished")
