"""PyAMQ Producer

This module represents an asynchronous activeMQ Producer written in Python.

Author: Thomas Low <thomas.low@cern.ch>
"""

import sys, time, threading, logging
from lib import brokermgr, util

log = logging.getLogger('pyamq.producer')

class Producer:
	"""Represents an asynchronous activeMQ Producer."""
	
	def __init__(self, broker, maxbroker=1, random=0):
		"""Initializes the publisher.
		
		BROKER is a list of broker URIs (e.g. stomp://host:port/?timeout=5)
		MAXBROKER is the number of simultanious connections
		RANDOM==0 means that broker will be choosen sequientially from the list.
		"""
		self.alive = 1
		self.event = threading.Event()
		self.brokermgr = brokermgr.BrokerManager(util.getUrlsAsDicts(broker), 
							maxbroker, random, self._event_handler)
		self.brokermgr.start()
		
	
	def __del__(self):
		self.stop(0)
		
		
	def _event_handler(self, event, data=None):
		if event == 'MESSAGE_SENT':
			self.event.set()
		elif event == 'EXIT':
			self.alive = 0
			self.event.set()
	

	def sendMessage(self, message):
		"""Sends the "message" and returns an identifier for it."""
		self.brokermgr.sendMessage(message)
		pass
		
	
	def allMessagesSent(self):
		"""Returns True if all messages are already sent."""
		return len(self.brokermgr.getPendingMessages()) == 0
	
	
	def wait(self, timeout):
		"""Blocks until all messages has been sent or the timeout exceeds or
		the publisher stops. Returns the tuple (messages not sent,
		messages sent) ."""
		endtime = time.time()+timeout
		pending = self.brokermgr.getPendingMessages()
		while endtime - time.time() > 0 and pending and self.alive:
			self.event.wait(endtime - time.time())
			self.event.clear()
			pending = self.brokermgr.getPendingMessages()
		return (pending, self.brokermgr.getSentMessages())
	
	
	def stop(self, timeout=0.5):
		"""Stops the publisher within maximal "timeout" seconds."""
		self.brokermgr.stop(timeout)
		self.alive = 0
		self.event.set()
		return (self.brokermgr.getPendingMessages(), 
				self.brokermgr.getSentMessages())
		
	
			

"""Test the publisher if not imported"""
if __name__ == '__main__':
	#config logging
	logging.getLogger('').setLevel(logging.DEBUG)
	console = logging.StreamHandler(sys.stdout)
	console.setLevel(logging.DEBUG)
	console.setFormatter(logging.Formatter(
		'%(relativeCreated)8d %(levelname)-8s: %(message)s (%(name)s)'
	))
	logging.getLogger('').addHandler(console)
	#config logging
	
	broker 	= ["http://gridmsg001.cern.ch/message"]
	dest	= "/topic/testtopic"
	
	log.info("program started")
	p = Producer(broker)
	for i in range(1,11): p.sendMessage({'header':{'destination': dest},'body':str(i)})
	p.wait(5)
	ns, s = p.stop()
	log.info("%d out of %d messages not sent" % (len(ns),len(ns)+len(s)))
	log.info("program stopped")
